'''
Software Carpentry EN.540.635
Lab 9 - Conjugate Gradient
'''
import numpy as np


def f1(x1, x2):
    '''
    Some arbitrary quadratic function.

    **Parameters**
        x1: *float*
        x2: *float*

    **Returns**
        y: *float*
    '''
    return (x1 - 2.0) ** 2 + (x2 - 3.0) ** 2


def g1(x1, x2):
    '''
    The gradient of f1.

    **Parameters**
        x1: *float*
        x2: *float*

    **Returns**
        y: *list, float*
    '''
    return [2.0 * (x1 - 2.0), 2.0 * (x2 - 3.0)]


def f2(x1):
    '''
    Some arbitrary trig function.

    **Parameters**

        x1: *float*

    **Returns**

        y: *float*
    '''
    return np.sin(3 * x1) + 0.05 * x1 ** 2


def g2(x1):
    '''
    The gradient of f2.

    **Parameters**
        x1: *float*

    **Returns**
        y: *float*
    '''
    return 3 * np.cos(3 * x1) + 0.1 * x1


def ensure_array_like(value):
    '''
    This function simply ensures that an input is an array/list.

    **Parameters**
        value: *...*
            Some object that should be array like.

    **Returns**
        value: *list or array like*
            The value now has a length!
    '''
    if not hasattr(value, "__len__"):
        value = [value]
    return value


def steepest_descent(func, grad, x0, MAXITER=1000000, step_size=0.001,
                     eps=1E-6):
    '''
    The steepest descent function, used to minimize some function f with
    its associated gradient g.

    **Parameters**
        func: *function*
            This is the function we want to minimize.
        grad: *function*
            This is the gradient of func.
        x0: *list, float*
            This is the starting coordinate for our optimizer.
        MAXITER: *int, optional*
            The maximum iteration our optimizer may run for.
        step_size: *float, optional*
            The scalar applied to the steepest descent step.
        eps: *float, optional*
            The tolerance for the gradient to end the Steepest Descent.

    **Returns**
        xopt: *list, float*
            The x at the minimum of func.
        fopt: *float*
            The function evaluation at xopt.
    '''
    # Handle if we input a value instead of an array
    x0 = np.array(ensure_array_like(x0))
    # Initialize flags
    GRAD_CONVERGED = False
    MAXITER_REACHED = False
    # Initialize x variable
    x = x0[:]
    for i in range(MAXITER):
        # Calculate gradient
        gradient = np.array(ensure_array_like(grad(*x)))
        # Check if gradient has converged to 0
        GRAD_CONVERGED = all([abs(g) < eps for g in gradient])
        if GRAD_CONVERGED:
            break
        # Calculate step
        step = -1.0 * step_size * gradient
        # Update x variable
        x += step
    # Define boolean for if we reached the maximum number of iterations
    MAXITER_REACHED = i == MAXITER - 1
    # Benchmarking print statements
    if GRAD_CONVERGED:
        print("\tGradient converged below %s!" % str(eps))
        print("\tIterations to Convergence: %d" % i)
    elif MAXITER_REACHED:
        print("\tFunction has not converged yet, maxiter was reached.")
    else:
        raise Exception("Error with flags.")
    return x, func(*x)


def conjugate_gradient(func, grad, x0, MAXITER=1000000, step_size=0.001,
                       eps=1E-6):
    '''
    The conjugate gradient function, used to minimize some function f with
    its associated gradient g.

    **Parameters**
        func: *function*
            This is the function we want to minimize.
        grad: *function*
            This is the gradient of func.
        x0: *list, float*
            This is the starting coordinate for our optimizer.
        MAXITER: *int, optional*
            The maximum iteration our optimizer may run for.
        step_size: *float, optional*
            The scalar applied to the steepest descent step.
        eps: *float, optional*
            The tolerance for the gradient to end the Steepest Descent.

    **Returns**
        xopt: *list, float*
            The x at the minimum of func.
        fopt: *float*
            The function evaluation at xopt.
    '''
    raise Exception("YOUR CODE GOES HERE")


def calculate_R2(xvals, y_obs, m, b):
    '''
    The R2 value for some given sampled points and slope/y-intercept.

    **Parameters**
        xvals: *list, float*
            A list of x values we sampled.
        y_obs: *list, float*
            A list of the sampled y values corresponding to the given
            x values.
        m: *float*
            A slope
        b: *float*
            A y-intercept.

    **Returns**
        R2: *float*
            The R2 value.

    **References**
        - https://en.wikipedia.org/wiki/Coefficient_of_determination
    '''
    y_mean = sum(y_obs) / len(y_obs)
    total_sum_of_errors = sum([
        (y - y_mean) ** 2 for y in y_obs
    ])
    residual_sum_of_errors = sum([
        (y - (m * x + b)) ** 2 for x, y in zip(xvals, y_obs)
    ])
    return 1.0 - residual_sum_of_errors / total_sum_of_errors


if __name__ == '__main__':
    ##########################################################################
    # STEP 1 - Implement your own Conjugate Gradient
    ##########################################################################
    x0 = [200.5, 13.5]
    print("-" * 79)
    print("\nTesting Quadratic Equation...\n")
    print("\tTesting Steepest Descent")
    xopt, fopt = steepest_descent(f1, g1, x0)
    print("\t\tf(%s) = %s" % (str(xopt), str(fopt)))
    print("\n\tTesting Conjugate Gradient")
    xopt, fopt = conjugate_gradient(f1, g1, x0)
    print("\t\tf(%s) = %s" % (str(xopt), str(fopt)))
    print("\n" + "-" * 79)

    x0 = 0.3
    print("\nTesting Trig Equation...\n")
    print("\tTesting Steepest Descent")
    xopt, fopt = steepest_descent(f2, g2, x0)
    print("\t\tf(%s) = %s" % (str(xopt), str(fopt)))
    print("\n\tTesting Conjugate Gradient")
    xopt, fopt = conjugate_gradient(f1, g1, x0)
    print("\t\tf(%s) = %s" % (str(xopt), str(fopt)))
    print("\n" + "-" * 79)

    ##########################################################################
    # STEP 2 - Implement Scipy
    ##########################################################################

    raise Exception("YOUR CODE GOES HERE")

    ##########################################################################
    # STEP 3 - Fit the line
    ##########################################################################

    # Class handout data
    x_vals = [0.00, 0.50, 1.00, 1.50, 2.00, 2.50, 3.00, 3.50, 4.00,
              4.50, 5.00, 5.50, 6.00, 6.50, 7.00, 7.50, 8.00, 8.50,
              9.00, 9.50]
    y_obs = [1.39, 0.25, -0.65, -2.38, -2.59, -5.03, -5.72, -6.28,
             -8.32, -9.49, -9.83, -12.01, -12.32, -13.21, -14.50,
             -15.51, -17.80, -18.41, -19.87, -21.31]

    raise Exception("YOUR CODE GOES HERE")
